﻿using System;
using System.Text;
using System.Diagnostics;
using UnityEngine;
using UnityEngine.UI;
using SimpleJSON;
using MacGruber;

// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/
// 
// Dependency:
// MacGruber.LogicBricks.12:/Custom/Scripts/MacGruber/LogicBricks/Internal/MacGruber_Utils.cs
// 
// Known bugs/issues:
// -> value for "Peak drop speed"-slider is to small to be visible
//
// Plugin usage examples:
// Light > LightSource > intensity/color/range
// Person > JawControl > targetRotationX (for basic 'lipsync')
// Person > BreastControl > spring or mass (bounce breast to beat)
// Person > AnimationControl > AnimatorSpeed
// anyMesh (AptSpeaker) > scale > scale 
// CoreControl > GlobalLighting > camExposure
namespace Sally {

	public class SoundFX : MVRScript {

		// PERFORMANCE CODE BEGIN
		private static double[] avgUpdateTimes = new double[80];
		private static double avgUpdateTime = 0;
		// PERFORMANCE CODE END

		private const int SAMPLE_SIZE = 512; // size requested from Unity
		private const int SAMPLE_SIZE_FAST = 128; // size actually used for Plugin
		public static float[] samples = new float[SAMPLE_SIZE]; // Unity Audio Spectrum Data
		public static float[] freqBands = new float[8]; // 8 frequency bands from first 128 samples
		public static float[] freqBandsAmp = new float[8];
		private static float samplefreq = 0f;
		private static float samples_sum = 0f;
		private const float SILENCE_THRESHOLD = 0.0055f;
		public static float[] peaks = new float[SAMPLE_SIZE_FAST];
		private static float[] triggerSource = null;
		private static float[] peakSpeeds = new float[SAMPLE_SIZE_FAST];
		private const float PEAK_DROP_SPEED_DEFAULT = 0.00025f;
		private static float peakDropSpeed = PEAK_DROP_SPEED_DEFAULT;
		private const float PEAK_GRAVITY_DEFAULT = 2.5f;
		private static float peakGravity = PEAK_GRAVITY_DEFAULT;
		public static float[] sin2x; // look up table for 2x positive sinus waves
		private static float deltaTimeDrawSinusLUT = 0f;

		private static Spectrum spectrum;
		private static GameObject goSpectrum;

		//JSONStorableString jsInfoL;
		private static UIDynamicTextField dText;
		private static StringBuilder stringBuilderInfo;
		private static int stringBuilderRemoveIndex = 0;
		private static JSONStorableString jsInfoR;
		private static JSONStorableBool jbFake;
		private static JSONStorableBool jbPeakTriggers;
		private static JSONStorableBool jbZoomSpectrum;
		private static JSONStorableBool jbShowBands;
		private static JSONStorableBool jbMicrophone;
		private static JSONStorableFloat jfAmplification;
		private static JSONStorableFloat jfPeakFallSpeed;
		private static JSONStorableFloat jfPeakGravity;

		// TRIGGER 1
		private static FloatTrigger fTrigger1;
		private static JSONStorableFloat jfTrigger1Freq;
		private static int sampleIndexTrigger1 = 1;
		private static JSONStorableBool jbSmoothTrigger1;
		private static float smoothTrigger1 = 0f;
		public static Color trigger1Color = new Color(1f, 0f, 0f, 1.0f);
		private static string trigger1ColorHex = ColorUtility.ToHtmlStringRGB(trigger1Color);

		// TRIGGER 2
		private static FloatTrigger fTrigger2;
		private static JSONStorableFloat jfTrigger2Freq;
		private static int sampleIndexTrigger2 = 15;
		private static JSONStorableBool jbSmoothTrigger2;
		private static float smoothTrigger2 = 0f;
		public static Color trigger2Color = new Color(0f, 0.4f, 1f, 1.0f);
		private static string trigger2ColorHex = ColorUtility.ToHtmlStringRGB(trigger2Color);

		public override void Init() {
			//SuperController.LogMessage("SallySoundFX Init()");
			base.Init();
			try {
				// TODO: Prevent Plugin being loaded multiple times to avoid wrong usage and high CPU load
				// TODO: Help button
				Mic.SelectByID(0);

				samplefreq = (float)AudioSettings.outputSampleRate / SAMPLE_SIZE * 0.5f;
				sin2x = LUT.CreateSinWave(128, 2, true);
				freqBandsAmp[0] = 1.00f;
				freqBandsAmp[1] = 1.00f;
				freqBandsAmp[2] = 1.10f;
				freqBandsAmp[3] = 2.80f;
				freqBandsAmp[4] = 3.50f;
				freqBandsAmp[5] = 8.00f;
				freqBandsAmp[6] = 14.00f;
				freqBandsAmp[7] = 20.00f;

				//////////////////////////////////////////////////////////////////////////////////
				// TOP
				//////////////////////////////////////////////////////////////////////////////////
				JSONStorableString jsInfoL = new JSONStorableString("SallySoundFX", "<size=40><b><color=#FF60E0>Sally</color><color=#D080F0>Sound</color><color=#A0E0FF>FX</color></b></size>" +
					"                                  <color=#FFFFFF>Trigger effects from audio.</color>" +
					"\n                                                                    <color=#FFFFFF20>Created by Sally Whitemane." +
					"\n                                                                    Licensed under CC BY SA." +
					"\n                                                                    Using MacGruber LogicBricks.</color>");
				dText = CreateTextField(jsInfoL, false);
				dText.height = 340f;
				Image imageBG = dText.GetComponentInChildren<Image>();
				imageBG.color = Color.black;

				goSpectrum = new GameObject("Spectrum");
				goSpectrum.transform.SetParent(dText.transform, false);
				spectrum = goSpectrum.AddComponent<Spectrum>();
				spectrum.transform.position = new Vector2(-515, -730);
				spectrum.sampleIndexTrigger1 = sampleIndexTrigger1;
				spectrum.sampleIndexTrigger2 = sampleIndexTrigger2;

				CreateSpacer(false);
				CreateSpacer(false);
				CreateSpacer(false);
				CreateSpacer(true);
				CreateSpacer(true);
				CreateSpacer(true);

				//////////////////////////////////////////////////////////////////////////////////
				// LEFT
				//////////////////////////////////////////////////////////////////////////////////
				stringBuilderInfo = new StringBuilder();
				stringBuilderInfo.Append("Unity sample rate: ").Append(AudioSettings.outputSampleRate).Append(" Hz ").AppendLine();
				stringBuilderInfo.Append("Plugin / Unity samples: ").Append(SAMPLE_SIZE_FAST).Append(" / ").Append(SAMPLE_SIZE).AppendLine();
				stringBuilderInfo.Append("Plugin freq. range: ").Append(samplefreq).Append(" - ").Append(samplefreq * SAMPLE_SIZE_FAST).Append(" Hz").AppendLine();
				stringBuilderRemoveIndex = stringBuilderInfo.Length;
				jsInfoR = Utils.SetupInfoText(this, stringBuilderInfo.ToString(), 230f, false);

				jbPeakTriggers = Utils.SetupToggle(this, "Use Peak values for triggers", false, false);

				// TRIGGER 1
				fTrigger1 = new FloatTrigger(this, "FloatTrigger1");
				Utils.SetupButton(this, "<color=#" + trigger1ColorHex + ">Trigger 1 Actions</color>", fTrigger1.OpenPanel, false);
				jbSmoothTrigger1 = Utils.SetupToggle(this, "<color=#" + trigger1ColorHex + ">Trigger 1 smooth spikes</color>", true, false);
				jfTrigger1Freq = Utils.SetupSliderFloat(this, "<color=#" + trigger1ColorHex + ">Trigger 1 frequency in Hz</color>", samplefreq * 2, samplefreq, samplefreq * SAMPLE_SIZE_FAST, false);
				jfTrigger1Freq.setCallbackFunction += SetTrigger1Frequency;

				// TRIGGER 2
				fTrigger2 = new FloatTrigger(this, "FloatTrigger2");
				Utils.SetupButton(this, "<color=#" + trigger2ColorHex + ">Trigger 2 Actions</color>", fTrigger2.OpenPanel, false);
				jbSmoothTrigger2 = Utils.SetupToggle(this, "<color=#" + trigger2ColorHex + ">Trigger 2 smooth spikes</color>", true, false);
				jfTrigger2Freq = Utils.SetupSliderFloat(this, "<color=#" + trigger2ColorHex + ">Trigger 2 frequency in Hz</color>", samplefreq * 16, samplefreq, samplefreq * SAMPLE_SIZE_FAST, false);
				jfTrigger2Freq.setCallbackFunction += SetTrigger2Frequency;

				//////////////////////////////////////////////////////////////////////////////////
				// RIGHT
				//////////////////////////////////////////////////////////////////////////////////
				Utils.SetupInfoText(this, "Easiest to sync is bass at ca. 100 Hz.\nTrigger values cannot be > 1.\n\nLike the free plugin? Support:\npatreon.com/sallyw\npatreon.com/MacGruber_Laboratory\nNo money? Rate or give feedback!", 230.0f, true);

				jbZoomSpectrum = Utils.SetupToggle(this, "Zoom Spectrum to 64 samples", true, true);
				jbFake = Utils.SetupToggle(this, "Test wave if silent", true, true);
				jbZoomSpectrum.setCallbackFunction += SetSpectrumSize;
				//jbShowBands = Utils.SetupToggle(this, "Show frequency bands", false, true);
				//jbShowBands.setCallbackFunction += ToggleShowFrequencyBands;
				jbMicrophone = Utils.SetupToggle(this, "Microphone", false, true);
				jbMicrophone.setCallbackFunction += ToggleMicrophone;

				jfAmplification = Utils.SetupSliderFloat(this, "Amplification", 20f, 1f, 200f, true);
				jfPeakFallSpeed = Utils.SetupSliderFloat(this, "Peak drop speed", PEAK_DROP_SPEED_DEFAULT, PEAK_DROP_SPEED_DEFAULT - 0.0002f, PEAK_DROP_SPEED_DEFAULT + 0.001f, true);
				jfPeakFallSpeed.setCallbackFunction += SetPeakFallSpeed;
				jfPeakGravity = Utils.SetupSliderFloat(this, "Peak gravity", PEAK_GRAVITY_DEFAULT, 0f, 10f, true);
				jfPeakGravity.setCallbackFunction += SetPeakGravity;

				// required for MacGruber_utils
				SimpleTriggerHandler.LoadAssets();
			}
			catch (Exception e) {
				SuperController.LogError("Exception: " + e);
			}
		}

		public override void InitUI() {
			//SuperController.LogMessage("SallySoundFX InitUI()");
			base.InitUI();
			if (UITransform == null)
				return;
			var scriptUI = UITransform.GetComponentInChildren<MVRScriptUI>();
			dText.transform.SetParent(scriptUI.fullWidthUIContent.transform, false);
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[fTrigger1.Name] = fTrigger1.GetJSON(base.subScenePrefix);
				jc[fTrigger2.Name] = fTrigger2.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				fTrigger1.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				fTrigger2.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		// Start is called once before Update or FixedUpdate is called and after Init()
		void Start() {
			//SuperController.LogMessage("SallySoundFX Start()");
			try	{
			}
			catch (Exception e) {
				SuperController.LogError("Exception: " + e);
			}
		}

		// Update is called with each rendered frame by Unity
		void Update()
		{
			try
			{
				// PERFORMANCE CODE BEGIN
				Stopwatch sw = new Stopwatch();
				sw.Start();
				// PERFORMANCE CODE END

				AudioListener.GetSpectrumData(samples, 0, FFTWindow.Blackman); // slow
				samples_sum = 0f;
				for (int i = 0; i < SAMPLE_SIZE_FAST; i++)
				{
					samples[i] = samples[i] * jfAmplification.val; // Amplify
					if (samples[i] > 1f) // Amp Limit
						samples[i] = 1f;
					samples_sum += samples[i]; // to detect silence
				}

				// No audio? Create fake sinus test signal. Shift generated sinus LUT (look up table).
				if (jbFake.val && samples_sum < SILENCE_THRESHOLD)
				{
					// limit the array shift to 30 times per frames (or else it's to fast)
					deltaTimeDrawSinusLUT += Time.deltaTime;
					if (deltaTimeDrawSinusLUT > 1f / 30f) 
					{
						deltaTimeDrawSinusLUT = deltaTimeDrawSinusLUT - 1f / 30f;
						float startToEnd = sin2x[0]; // store old start value
						Array.Copy(sin2x, 1, sin2x, 0, sin2x.Length - 1); // shift array
						sin2x[sin2x.Length - 1] = startToEnd; // move old start value to end
					}
					for (int i = 0; i < SAMPLE_SIZE_FAST; i++)
						samples[i] = sin2x[i]; // LUT uses custom positive sinus wave
				}

				UpdateFrequencyBands();

				// Peaks
				for (int i = 0; i < SAMPLE_SIZE_FAST; i++)
				{
					peakSpeeds[i] *= 1f + Time.deltaTime * peakGravity;
					peaks[i] -= peakSpeeds[i] * Time.deltaTime * 1000f;
					if (peaks[i] < samples[i])
					{
						// new peak value, set peak to current value and reset peakSpeed
						peaks[i] = samples[i];
						peakSpeeds[i] = peakDropSpeed;
					}
				}

				// No audio? Create fake sinus test signal. Slower code because Mathf.Sin() is used
				//if (jbFake.val && samples_fast_sum < 0.01f)
				//	for (i = 0; i < SAMPLE_SIZE_FAST; i++)
				//		samples_fast[i] = 0.5f + 0.5f * Mathf.Sin(Time.time + i * 0.2f);

				// SET SOURCE FOR TRIGGERS
				if (jbPeakTriggers.val)
					triggerSource = peaks;
				else
					triggerSource = samples;

				// TRIGGER 1 SMOOTH
				if (jbSmoothTrigger1.val)
				{
					fTrigger1.Trigger((triggerSource[sampleIndexTrigger1] + smoothTrigger1) * 0.5f);
					smoothTrigger1 = triggerSource[sampleIndexTrigger1];
				}
				else
					fTrigger1.Trigger(triggerSource[sampleIndexTrigger1]);

				// TRIGGER 2 SMOOTH
				if (jbSmoothTrigger2.val)
				{
					fTrigger2.Trigger((triggerSource[sampleIndexTrigger2] + smoothTrigger2) * 0.5f);
					smoothTrigger2 = triggerSource[sampleIndexTrigger2];
				}
				else
					fTrigger2.Trigger(triggerSource[sampleIndexTrigger2]);

				spectrum.UpdateSamples();

				// PERFORMANCE CODE BEGIN
				sw.Stop();
				avgUpdateTimes[avgUpdateTimes.Length - 1] = sw.Elapsed.TotalMilliseconds;
				Array.Copy(avgUpdateTimes, 1, avgUpdateTimes, 0, avgUpdateTimes.Length - 1); // shift array
				avgUpdateTime = 0;
				for (int i = 0; i < avgUpdateTimes.Length; i++)
					avgUpdateTime += avgUpdateTimes[i];
				avgUpdateTime = avgUpdateTime / avgUpdateTimes.Length;
				// PERFORMANCE CODE END

				jsInfoR.SetVal(GetUpdatedInfoText());
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception: " + e);
			}
		}

		// FixedUpdate is called with each physics simulation frame by Unity
		void FixedUpdate()
		{
		}

		// OnDestroy is where you should put any cleanup
		// if you registered objects to supercontroller or atom, you should unregister them here
		void OnDestroy()
		{
			fTrigger1.Remove();
			fTrigger2.Remove();
			Destroy(goSpectrum);
		}

		//////////////////////////////////////////////////////////////////////////////////
		// GUI Callbacks
		//////////////////////////////////////////////////////////////////////////////////
		private static void SetPeakFallSpeed(float v) {peakDropSpeed = v;}
		private static void SetPeakGravity(float v) {peakGravity = v;}
		private static void SetSpectrumSize(bool b) {if (b)	spectrum.size = 64;	else spectrum.size = 128;}
		private static void SetTrigger1Frequency(float f) {
			f = SnapToSampleFrequency(f);
			jfTrigger1Freq.val = f;
			sampleIndexTrigger1 = (int)(f / samplefreq) -1;
			spectrum.sampleIndexTrigger1 = sampleIndexTrigger1;
		}
		private static void SetTrigger2Frequency(float f) {
			f = SnapToSampleFrequency(f);
			jfTrigger2Freq.val = f;
			sampleIndexTrigger2 = (int)(f / samplefreq) - 1;
			spectrum.sampleIndexTrigger2 = sampleIndexTrigger2;
		}
		private static void ToggleShowFrequencyBands(bool b) {spectrum.showFrequencyBands = b;}
		private static void ToggleMicrophone(bool b) {if (b) Mic.On(); else	Mic.Off();}

		//////////////////////////////////////////////////////////////////////////////////
		// Helper functions
		//////////////////////////////////////////////////////////////////////////////////
		private static float SnapToSampleFrequency(float v) {
			return (float)Math.Round(v / samplefreq) * samplefreq;
		}

		private static String GetUpdatedInfoText() {
			stringBuilderInfo.Remove(stringBuilderRemoveIndex, stringBuilderInfo.Length - stringBuilderRemoveIndex);
			// PERFORMANCE CODE BEGIN
			stringBuilderInfo.Append("Average update time: ");
			String colorString = "<color=#00FF00>";
			if (avgUpdateTime >= 0.25)
				colorString = "<color=#FFFF00>";
			else if (avgUpdateTime >= 0.5)
				colorString = "<color=#FF0000>";
			stringBuilderInfo.Append(colorString).Append(avgUpdateTime.ToString("0.0000 ms")).Append("</color>");
			stringBuilderInfo.AppendLine();
			// PERFORMANCE CODE END
			stringBuilderInfo.Append("<b><color=#").Append(trigger1ColorHex).Append(">").Append(samples[sampleIndexTrigger1].ToString("0.000")).Append("</color>    <color=#").Append(trigger2ColorHex).Append(">").Append(samples[sampleIndexTrigger2].ToString("0.000</color></b>"));
			return stringBuilderInfo.ToString();
		}

		private static void UpdateFrequencyBands() {
			// band 0: 1  sample  index 0-0
			// band 1: 1  sample  index 1-1
			// band 2: 2  sampels index 2-3
			// band 3: 4  samples index 4-7
			// band 4: 8  samples index 8-15
			// band 5: 16 samples index 16-31
			// band 6: 32 samples index 32-63
			// band 7: 64 samples index 64-127

			int i;

			// band 0
			freqBands[0] = samples[0];
			// band 1
			freqBands[1] = samples[1];
			// band 2
			freqBands[2] = (samples[2] + samples[3]) * 0.5f;
			// band 3
			freqBands[3] = (samples[4] + samples[5] + samples[6] + samples[7]) * 0.25f;

			// band 4
			for (i = 8; i < 16; i++)
				freqBands[4] += samples[i];
			freqBands[4] *= 0.125f;    // avoid slow divide by 8 => multiply by 0.125f
			// band 5
			for (i = 16; i < 32; i++)
				freqBands[5] += samples[i];
			freqBands[5] *= 0.0625f;   // avoid slow divide by 16 => multiply by 0.0625f
			// band 6
			for (i = 32; i < 64; i++)
				freqBands[6] += samples[i];
			freqBands[6] *= 0.03125f;  // avoid slow divide by 32 => multiply by 0.03125f
			// band 7
			for (i = 64; i < 127; i++)
				freqBands[7] += samples[i];
			freqBands[7] *= 0.015625f; // avoid slow divide by 64 => multiply by 0.015625f

			/*
			// BAND 4-8 SLOWER ALT CODE
			int band = 4;
			int sampleCount = 0;
			for (i = 8; i < SAMPLE_SIZE_FAST; i++)
			{
				if (i == 16 || i == 32 || i == 64)
                {
					freqBands[band] /= sampleCount;
					sampleCount = 0;
					band++;
				}
				freqBands[band] += samples[i];
				sampleCount++;
			}
			freqBands[band] /= sampleCount;
			*/

			// Amplify
			for (i = 0; i < 8; i++)
            {
                freqBands[i] *= freqBandsAmp[i];
                if (freqBands[i] > 1f)
                    freqBands[i] = 1f;
            }
		}
    }
}