﻿using UnityEngine;
using UnityEngine.UI;

// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/
namespace Sally
{
	public class Spectrum : Graphic
	{
		private float _width = 1024;// 128 x 8 bar width = 1024
		private float _height = 280;
		private float[] samples;
		private float[] peaks;
		private float[] bands;
		public int sampleIndexTrigger1 = 1;
		public int sampleIndexTrigger2 = 15;
		public int size = 64;
		public bool showFrequencyBands = false;

		public float Width
		{
			get { return _width; }
			set { _width = value; SetVerticesDirty(); }
		}

		public float Height
		{
			get { return _height; }
			set { _height = value; SetVerticesDirty(); }
		}

		protected override void OnPopulateMesh(VertexHelper vh)
		{
			vh.Clear();
			//rectTransform.sizeDelta = new Vector2(_width, _height);
			//rectTransform.pivot = Vector2.zero;
			UpdateBars(vh);
		}

		private void UpdateBars(VertexHelper vh)
		{
			UIVertex vertex = UIVertex.simpleVert;

			int vIndex = 0;
			if (showFrequencyBands)
            {
				float barWidth = _width / bands.Length;
				for (int i = 0; i < bands.Length; i++)
				{
					vertex.color = new Color(1f * bands[i], 1f - bands[i], 1f, 1.0f);
					vertex.position = new Vector3(i * barWidth, 0);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth, bands[i] * _height);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth + barWidth, bands[i] * _height);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth + barWidth, 0);
					vh.AddVert(vertex);

					vIndex = i * 4;
					vh.AddTriangle(vIndex, vIndex + 1, vIndex + 2);     // bar triangle 1
					vh.AddTriangle(vIndex + 2, vIndex + 3, vIndex);     // bar triangle 2
				}
			}
			else
            {
				float barWidth = _width / size;
				for (int i = 0; i < size; i++)
				{
					vertex.color = new Color(1f * samples[i], 1f - samples[i], 1f, 1.0f);
					if (i == sampleIndexTrigger1)
						vertex.color = SoundFX.trigger1Color;
					if (i == sampleIndexTrigger2)
						vertex.color = SoundFX.trigger2Color;
					vertex.position = new Vector3(i * barWidth, 0);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth, samples[i] * _height);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth + barWidth, samples[i] * _height);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth + barWidth, 0);
					vh.AddVert(vertex);

					// peaks
					vertex.color = Color.white;
					vertex.position = new Vector3(i * barWidth, peaks[i] * _height);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth + barWidth * 0.5f, peaks[i] * _height + 8f);
					vh.AddVert(vertex);
					vertex.position = new Vector3(i * barWidth + barWidth, peaks[i] * _height);
					vh.AddVert(vertex);

					vIndex = i * 7;
					vh.AddTriangle(vIndex, vIndex + 1, vIndex + 2);     // bar triangle 1
					vh.AddTriangle(vIndex + 2, vIndex + 3, vIndex);     // bar triangle 2
					vh.AddTriangle(vIndex + 4, vIndex + 5, vIndex + 6); // peak triangle
				}
			}
		}

		public void UpdateSamples()
		{
			this.samples = SoundFX.samples;
			this.peaks = SoundFX.peaks;
			this.bands = SoundFX.freqBands;
			SetVerticesDirty();
		}
	}
}